/**
 * JASMINe VMMapi: JASMINe Virtual Machine Management API
 * Copyright (C) 2009 France Telecom R&D
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: VirtManagerAgent.java 5514 2009-10-13 16:41:18Z dangtran $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.vmm.agent.main;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.util.Properties;

import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.log4j.Logger;
import org.ow2.jasmine.vmm.agent.domain.Domain;
import org.ow2.jasmine.vmm.agent.domain.DomainBuilder;
import org.ow2.jasmine.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.jasmine.vmm.api.DomainMXBean;
import org.ow2.jasmine.vmm.api.VMMAgentMXBean;

/**
 * VirtManagerAgent implements the VMMAgentMXBean interface and is the entry
 * point of the agent. VirtManagerAgent is a singleton. VirtManagerAgent
 * expects:
 * <ul>
 * <li>the vmm.home property to be set to the directory where the VMM agent
 * distribution is installed
 * <li>the etc/agent.properties file to contains agent properties
 * </ul>
 */
public class VirtManagerAgent extends AgentCommon implements VMMAgentMXBean {
    private static Logger logger = Logger.getLogger(VirtManagerAgent.class);

    public final static String vmmHome = System.getProperty("vmm.home");

    static VirtManagerAgent instance;

    Domain rootDomain;

    /*
     * @see org.ow2.jasmine.vmm.api.VMMAgentMXBean#getRootDomain()
     */
    public DomainMXBean getRootDomain() {
        return this.rootDomain;
    }

    public static void main(final String[] args) {
        if (VirtManagerAgent.vmmHome == null) {
            VirtManagerAgent.logger.error("vmm.home system property must be set");
            System.exit(1);
        }
        VirtManagerAgent.instance = new VirtManagerAgent(args);
    }

    private VirtManagerAgent(final String[] args) {
        VirtManagerAgent.instance = this;
        Properties properties = new Properties();
        String agentPropFileName = "etc" + File.separator + "agent.properties";
        try {
            properties.load(new FileInputStream(new File(VirtManagerAgent.vmmHome, agentPropFileName)));
        } catch (IOException ex) {
            VirtManagerAgent.logger.fatal("Cannot load VMM agent property file " + agentPropFileName, ex);
            System.exit(1);
        }

        ObjectName agentName = MBeanObjectNamer.makeAgentName();
        try {
            AgentCommon.mbs.registerMBean(this, agentName);
        } catch (Exception ex) {
            VirtManagerAgent.logger.fatal("Cannot register MXBean " + agentName);
            System.exit(1);
        }
        VirtManagerAgent.logger.info("Created VMMAgent MBean " + agentName);

        int port;
        try {
            port = Integer.parseInt(properties.getProperty("vmm.port"));
        } catch (NumberFormatException ex) {
            port = 9999;
        }

        String managedResourcesFileName = properties.getProperty("vmm.resourceFile", "managed-resources.xml");
        File managedResourcesFile = new File(VirtManagerAgent.vmmHome, "etc" + File.separator + managedResourcesFileName);

        this.rootDomain = DomainBuilder.createManagedResourcesFromXML(managedResourcesFile);
        if (this.rootDomain == null) {
            VirtManagerAgent.logger.fatal("Failed to load resource file");
            System.exit(1);
        }

        try {
            LocateRegistry.createRegistry(port);
        } catch (RemoteException ex) {
            VirtManagerAgent.logger.warn("Cannot launch RMI registry on port " + port + ", creating new one");
            try {
                LocateRegistry.createRegistry(port);
            } catch (RemoteException ex2) {
                VirtManagerAgent.logger.fatal("Cannot create RMI registry", ex);
                System.exit(1);
            }
        }
        try {
            JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://localhost:" + port + "/server");
            JMXConnectorServer cs = JMXConnectorServerFactory.newJMXConnectorServer(url, null, AgentCommon.mbs);
            VirtManagerAgent.logger.info("Starting RMI connector server with URL \"" + url + "\"");
            cs.start();
            VirtManagerAgent.logger.info("VMM agent ready");
        } catch (Exception ex) {
            VirtManagerAgent.logger.fatal("Failed to starte JMX connector server", ex);
            System.exit(1);
        }
    }

    /*
     * @see org.ow2.jasmine.vmm.api.VMMAgentMXBean#shutdown()
     */
    public void shutdown() {
        // we avoid to exit synchronously, otherwise
        // remote clients will fail when invoking this method
        new Thread(new Runnable() {
            public void run() {
                try {
                    Thread.sleep(1000);
                } catch (Exception ex) {
                }
                ;
                // TODO clean-up before bailing out
                VirtManagerAgent.logger.info("VMM agent terminated");
                System.exit(0);
            }
        }).run();
    }

}